"""Test scaling data"""

import numpy as np
import pytest

from nptdms import types
from nptdms.scaling import get_scaling


def test_unsupported_scaling_type():
    """Raw data is returned unscaled when the scaling type is unsupported.
    """

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "UnknownScaling"
    }
    scaling = get_scaling(properties, {}, {})

    assert scaling is None


def test_linear_scaling():
    """Test linear scaling"""

    data = StubTdmsData(np.array([1, 2, 3], dtype=np.dtype('int32')))
    expected_scaled_data = np.array([12.0, 14.0, 16.0])

    properties = {
        "NI_Scaling_Status": "unscaled",
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Linear",
        "NI_Scale[0]_Linear_Slope": 2.0,
        "NI_Scale[0]_Linear_Y_Intercept": 10.0
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.Int32, None) == np.dtype('float64')
    assert scaled_data.dtype == np.dtype('float64')
    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_polynomial_scaling():
    """Test polynomial scaling"""

    data = StubTdmsData(np.array([1, 2, 3], dtype=np.dtype('int32')))
    expected_scaled_data = np.array([16.0, 44.0, 112.0])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Polynomial",
        "NI_Scale[0]_Polynomial_Coefficients[0]": 10.0,
        "NI_Scale[0]_Polynomial_Coefficients[1]": 1.0,
        "NI_Scale[0]_Polynomial_Coefficients[2]": 2.0,
        "NI_Scale[0]_Polynomial_Coefficients[3]": 3.0,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.Int32, None) == np.dtype('float64')
    assert scaled_data.dtype == np.dtype('float64')
    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_polynomial_scaling_with_no_coefficients():
    """Test polynomial scaling when there are no coefficients, so data should be all zero
    """
    data = StubTdmsData(np.array([1.0, 2.0, 3.0]))
    expected_scaled_data = np.array([0.0, 0.0, 0.0])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Polynomial",
        "NI_Scale[0]_Polynomial_Coefficients_Size": 0
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_polynomial_scaling_with_3_coefficients():
    """Test polynomial scaling"""

    data = StubTdmsData(np.array([1, 2, 3], dtype=np.dtype('int32')))
    expected_scaled_data = np.array([13.0, 20.0, 31.0])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Polynomial",
        "NI_Scale[0]_Polynomial_Coefficients_Size": 3,
        "NI_Scale[0]_Polynomial_Coefficients[0]": 10.0,
        "NI_Scale[0]_Polynomial_Coefficients[1]": 1.0,
        "NI_Scale[0]_Polynomial_Coefficients[2]": 2.0,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


@pytest.mark.parametrize(
    "resistance_configuration,lead_resistance,expected_data",
    [
        (2, 0.0, [1256.89628, 1712.83429]),
        (2, 100.0, [557.6879004146, 882.7374139697]),
        (3, 0.0, [1256.89628, 1712.83429]),
        (3, 100.0, [882.7374139697, 1256.896275222]),
        (4, 0.0, [1256.89628, 1712.83429]),
        (4, 100.0, [1256.89628, 1712.83429]),
    ]
)
def test_rtd_scaling(resistance_configuration, lead_resistance, expected_data):
    """Test RTD scaling"""

    data = StubTdmsData(np.array([0.5, 0.6]))

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "RTD",
        "NI_Scale[0]_RTD_Current_Excitation": 0.001,
        "NI_Scale[0]_RTD_R0_Nominal_Resistance": 100.0,
        "NI_Scale[0]_RTD_A": 0.0039083,
        "NI_Scale[0]_RTD_B": -5.775e-07,
        "NI_Scale[0]_RTD_C": -4.183e-12,
        "NI_Scale[0]_RTD_Lead_Wire_Resistance": lead_resistance,
        "NI_Scale[0]_RTD_Resistance_Configuration": resistance_configuration,
        "NI_Scale[0]_RTD_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.DoubleFloat, None) == np.dtype('float64')
    np.testing.assert_almost_equal(scaled_data, expected_data, decimal=3)


def test_rtd_scaling_with_negative_temperature():
    """ Test RTD scaling with negative temperature values, which requires
        solving the full quartic Callendar-Van Dusen equation
    """
    data = StubTdmsData(np.array([
        0.08, 0.09, 0.095, 0.1, 0.12, 0.14, 0.16, 0.18, 0.2, 0.22, 0.24]))
    expected_data = np.array([
        -50.77114, -25.48835, -12.76894, -0., 51.56605, 103.94273,
        157.1695, 211.28915, 266.34819, 322.3973, 379.49189])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "RTD",
        "NI_Scale[0]_RTD_Current_Excitation": 0.001,
        "NI_Scale[0]_RTD_R0_Nominal_Resistance": 100.0,
        "NI_Scale[0]_RTD_A": 0.0039083,
        "NI_Scale[0]_RTD_B": -5.775e-07,
        "NI_Scale[0]_RTD_C": -4.183e-12,
        "NI_Scale[0]_RTD_Lead_Wire_Resistance": 0,
        "NI_Scale[0]_RTD_Resistance_Configuration": 2,
        "NI_Scale[0]_RTD_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.DoubleFloat, None) == np.dtype('float64')
    assert scaled_data.dtype == np.dtype('float64')
    np.testing.assert_almost_equal(scaled_data, expected_data, decimal=5)


def test_table_scaling():
    """Test table scaling"""

    data = StubTdmsData(np.array([0.5, 1.0, 1.5, 2.5, 3.0, 3.5]))
    expected_scaled_data = np.array([2.0, 2.0, 3.0, 6.0, 8.0, 8.0])

    # The scaled values are actually the range of inputs into the scaling,
    # which are mapped to the pre-scaled values. This makes no sense but
    # matches the behaviour of the Excel TDMS plugin.

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Table",
        "NI_Scale[0]_Table_Scaled_Values_Size": 3,
        "NI_Scale[0]_Table_Scaled_Values[0]": 1.0,
        "NI_Scale[0]_Table_Scaled_Values[1]": 2.0,
        "NI_Scale[0]_Table_Scaled_Values[2]": 3.0,
        "NI_Scale[0]_Table_Pre_Scaled_Values_Size": 3,
        "NI_Scale[0]_Table_Pre_Scaled_Values[0]": 2.0,
        "NI_Scale[0]_Table_Pre_Scaled_Values[1]": 4.0,
        "NI_Scale[0]_Table_Pre_Scaled_Values[2]": 8.0,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.DoubleFloat, None) == np.dtype('float64')
    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_add_scaling():
    """ Test scaling that adds two input scalings"""

    scaler_data = StubDaqmxData({
        0: np.array([1, 2, 3], dtype=np.dtype('int32')),
        1: np.array([2, 4, 6], dtype=np.dtype('uint32')),
    })
    expected_scaled_data = np.array([3.0, 6.0, 9.0])

    properties = {
        "NI_Number_Of_Scales": 3,
        "NI_Scale[2]_Scale_Type": "Add",
        "NI_Scale[2]_Add_Left_Operand_Input_Source": 0,
        "NI_Scale[2]_Add_Right_Operand_Input_Source": 1,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(scaler_data)

    assert scaling.get_dtype(None, {0: types.Int32, 1: types.Uint32}) == np.dtype('int64')
    assert scaled_data.dtype == np.dtype('int64')
    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_add_scaling_with_default_inputs():
    """ Test scaling that adds two input scalings"""

    data = StubTdmsData(np.array([1, 2, 3], dtype=np.dtype('int32')))
    expected_scaled_data = np.array([2, 4, 6])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Add",
        "NI_Scale[0]_Add_Left_Operand_Input_Source": 0xFFFFFFFF,
        "NI_Scale[0]_Add_Right_Operand_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.Int32, None) == np.dtype('int32')
    assert scaled_data.dtype == np.dtype('int32')
    np.testing.assert_equal(expected_scaled_data, scaled_data)


def test_subtract_scaling():
    """ Test scaling that subtracts an input scaling from another"""

    # This behaves the opposite to what you'd expect, the left operand
    # is subtracted from the right operand.
    scaler_data = StubDaqmxData({
        0: np.array([1, 2, 3], dtype=np.dtype('int32')),
        1: np.array([2, 4, 6], dtype=np.dtype('uint32')),
    })
    expected_scaled_data = np.array([1.0, 2.0, 3.0])

    properties = {
        "NI_Number_Of_Scales": 3,
        "NI_Scale[2]_Scale_Type": "Subtract",
        "NI_Scale[2]_Subtract_Left_Operand_Input_Source": 0,
        "NI_Scale[2]_Subtract_Right_Operand_Input_Source": 1,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(scaler_data)

    assert scaling.get_dtype(None, {0: types.Int32, 1: types.Uint32}) == np.dtype('int64')
    assert scaled_data.dtype == np.dtype('int64')
    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_thermocouple_scaling_voltage_to_temperature():
    """Test thermocouple scaling from a voltage in uV to temperature"""

    data = StubTdmsData(np.array([0.0, 10.0, 100.0, 1000.0]))
    expected_scaled_data = np.array([0.000000, 0.250843, 2.508899, 24.983648])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermocouple",
        "NI_Scale[0]_Thermocouple_Thermocouple_Type": 10073,
        "NI_Scale[0]_Thermocouple_Scaling_Direction": 0,
        "NI_Scale[0]_Thermocouple_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(
        scaled_data, expected_scaled_data, decimal=3)


def test_thermocouple_scaling_temperature_to_voltage():
    """Test thermocouple scaling from a temperature to voltage in uV"""

    data = StubTdmsData(np.array([0.0, 10.0, 50.0, 100.0]))
    expected_scaled_data = np.array([
        0.0, 396.8619078, 2023.0778862, 4096.2302187])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermocouple",
        "NI_Scale[0]_Thermocouple_Thermocouple_Type": 10073,
        "NI_Scale[0]_Thermocouple_Scaling_Direction": 1,
        "NI_Scale[0]_Thermocouple_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(
        scaled_data, expected_scaled_data, decimal=3)


def test_thermocouple_scaling_voltage_to_temperature_benchmark(benchmark):
    """Test thermocouple scaling from a voltage in uV to temperature"""

    data = StubTdmsData(np.tile(np.array([0.0, 10.0, 100.0, 1000.0]), 100))
    expected_scaled_data = np.tile(np.array([0.000000, 0.250843, 2.508899, 24.983648]), 100)

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermocouple",
        "NI_Scale[0]_Thermocouple_Thermocouple_Type": 10073,
        "NI_Scale[0]_Thermocouple_Scaling_Direction": 0,
        "NI_Scale[0]_Thermocouple_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = benchmark(scaling.scale, data)

    np.testing.assert_almost_equal(
        scaled_data, expected_scaled_data, decimal=3)


def test_thermocouple_scaling_temperature_to_voltage_benchmark(benchmark):
    """Test thermocouple scaling from a temperature to voltage in uV"""

    data = StubTdmsData(np.tile(np.array([0.0, 10.0, 50.0, 100.0]), 100))
    expected_scaled_data = np.tile(np.array([
        0.0, 396.8619078, 2023.0778862, 4096.2302187]), 100)

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermocouple",
        "NI_Scale[0]_Thermocouple_Thermocouple_Type": 10073,
        "NI_Scale[0]_Thermocouple_Scaling_Direction": 1,
        "NI_Scale[0]_Thermocouple_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = benchmark(scaling.scale, data)

    np.testing.assert_almost_equal(
        scaled_data, expected_scaled_data, decimal=3)


@pytest.mark.parametrize(
    "resistance_configuration,lead_resistance,expected_data",
    [
        (2, 0.0, [287.1495569816, 290.71633623, 294.4862276706]),
        (2, 100.0, [287.1495569816, 290.71633623, 294.4862276706]),
        (3, 0.0, [287.1495569816, 290.71633623, 294.4862276706]),
        (3, 100.0, [287.4248927942, 291.0482875767, 294.8892119392]),
        (4, 0.0, [287.1495569816, 290.71633623, 294.4862276706]),
        (4, 100.0, [287.1495569816, 290.71633623, 294.4862276706]),
    ]
)
def test_thermistor_scaling_with_voltage_excitation(
        resistance_configuration, lead_resistance, expected_data):
    data = StubTdmsData(np.array([1.1, 1.0, 0.9]))

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermistor",
        "NI_Scale[0]_Thermistor_Resistance_Configuration": resistance_configuration,
        "NI_Scale[0]_Thermistor_Excitation_Type": 10322,
        "NI_Scale[0]_Thermistor_Excitation_Value": 2.5,
        "NI_Scale[0]_Thermistor_R1_Reference_Resistance": 10000.0,
        "NI_Scale[0]_Thermistor_Lead_Wire_Resistance": lead_resistance,
        "NI_Scale[0]_Thermistor_A": 0.0012873851,
        "NI_Scale[0]_Thermistor_B": 0.00023575235,
        "NI_Scale[0]_Thermistor_C": 9.497806e-8,
        "NI_Scale[0]_Thermistor_Temperature_Offset": 1.0,
        "NI_Scale[0]_Thermistor_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_data)


@pytest.mark.parametrize(
    "resistance_configuration,lead_resistance,expected_data",
    [
        (2, 0.0, [335.5876272527, 338.303823856, 341.3530400858]),
        (2, 100.0, [341.3530400858, 344.8212218133, 348.831282405]),
        (3, 0.0, [335.5876272527, 338.303823856, 341.3530400858]),
        (3, 100.0, [338.303823856, 341.3530400858, 344.8212218133]),
        (4, 0.0, [335.5876272527, 338.303823856, 341.3530400858]),
        (4, 100.0, [335.5876272527, 338.303823856, 341.3530400858]),
    ]
)
def test_thermistor_scaling_with_current_excitation(
        resistance_configuration, lead_resistance, expected_data):
    data = StubTdmsData(np.array([1.1, 1.0, 0.9]))

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermistor",
        "NI_Scale[0]_Thermistor_Resistance_Configuration": resistance_configuration,
        "NI_Scale[0]_Thermistor_Excitation_Type": 10134,
        "NI_Scale[0]_Thermistor_Excitation_Value": 1.0e-3,
        "NI_Scale[0]_Thermistor_R1_Reference_Resistance": 0.0,
        "NI_Scale[0]_Thermistor_Lead_Wire_Resistance": lead_resistance,
        "NI_Scale[0]_Thermistor_A": 0.0012873851,
        "NI_Scale[0]_Thermistor_B": 0.00023575235,
        "NI_Scale[0]_Thermistor_C": 9.497806e-8,
        "NI_Scale[0]_Thermistor_Temperature_Offset": 1.0,
        "NI_Scale[0]_Thermistor_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_data)


def test_thermistor_scaling_with_invalid_excitation_type():
    data = StubTdmsData(np.array([1.1, 1.0, 0.9]))

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Thermistor",
        "NI_Scale[0]_Thermistor_Resistance_Configuration": 3,
        "NI_Scale[0]_Thermistor_Excitation_Type": 12345,
        "NI_Scale[0]_Thermistor_Excitation_Value": 2.5,
        "NI_Scale[0]_Thermistor_R1_Reference_Resistance": 10000.0,
        "NI_Scale[0]_Thermistor_Lead_Wire_Resistance": 0.0,
        "NI_Scale[0]_Thermistor_A": 0.0012873851,
        "NI_Scale[0]_Thermistor_B": 0.00023575235,
        "NI_Scale[0]_Thermistor_C": 9.497806e-8,
        "NI_Scale[0]_Thermistor_Temperature_Offset": 1.0,
        "NI_Scale[0]_Thermistor_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    with pytest.raises(ValueError) as exc_info:
        _ = scaling.scale(data)
    assert "Invalid excitation type: 12345" in str(exc_info.value)


_strain_voltage = [
        0.0068827, 0.0068036, 0.00688, 0.0068545, 0.0069104,
        0.0068033, 0.0068023, 0.0068316, 0.0067672, 0.00679]

_measured_strain = {
    10183: {
        'baseline': [
            -1.310990e-03, -1.295924e-03, -1.310476e-03, -1.305619e-03, -1.316267e-03,
            -1.295867e-03, -1.295676e-03, -1.301257e-03, -1.288990e-03, -1.293333e-03,
        ],
        'with_adjustment': [
            -1.472242e-03, -1.455322e-03, -1.471665e-03, -1.466210e-03, -1.478167e-03,
            -1.455258e-03, -1.455044e-03, -1.461312e-03, -1.447536e-03, -1.452413e-03,
        ],
        'with_initial_voltage': [
            -1.053848e-03, -1.038781e-03, -1.053333e-03, -1.048476e-03, -1.059124e-03,
            -1.038724e-03, -1.038533e-03, -1.044114e-03, -1.031848e-03, -1.036190e-03,
        ],
        'with_lead_resistance': [
            -1.310990e-03, -1.295924e-03, -1.310476e-03, -1.305619e-03, -1.316267e-03,
            -1.295867e-03, -1.295676e-03, -1.301257e-03, -1.288990e-03, -1.293333e-03,
        ],
        'with_all': [
            -1.183471e-03, -1.166551e-03, -1.182893e-03, -1.177439e-03, -1.189396e-03,
            -1.166487e-03, -1.166273e-03, -1.172540e-03, -1.158765e-03, -1.163642e-03,
        ],
    },
    10184: {
        'baseline': [
            -2.016908e-03, -1.993729e-03, -2.016117e-03, -2.008645e-03, -2.025026e-03,
            -1.993641e-03, -1.993348e-03, -2.001934e-03, -1.983062e-03, -1.989744e-03,
        ],
        'with_adjustment': [
            -2.264988e-03, -2.238958e-03, -2.264100e-03, -2.255708e-03, -2.274104e-03,
            -2.238859e-03, -2.238530e-03, -2.248172e-03, -2.226979e-03, -2.234482e-03,
        ],
        'with_initial_voltage': [
            -1.621304e-03, -1.598125e-03, -1.620513e-03, -1.613040e-03, -1.629421e-03,
            -1.598037e-03, -1.597744e-03, -1.606330e-03, -1.587458e-03, -1.594139e-03,
        ],
        'with_lead_resistance': [
            -2.016908e-03, -1.993729e-03, -2.016117e-03, -2.008645e-03, -2.025026e-03,
            -1.993641e-03, -1.993348e-03, -2.001934e-03, -1.983062e-03, -1.989744e-03,
        ],
        'with_all': [
            -1.820724e-03, -1.794694e-03, -1.819836e-03, -1.811444e-03, -1.829840e-03,
            -1.794595e-03, -1.794266e-03, -1.803908e-03, -1.782715e-03, -1.790218e-03,
        ],
    },
    10185: {
        'baseline': [
            -2.013923e-03, -1.990812e-03, -2.013134e-03, -2.005684e-03, -2.022016e-03,
            -1.990724e-03, -1.990432e-03, -1.998993e-03, -1.980176e-03, -1.986838e-03,
        ],
        'with_adjustment': [
            -2.261635e-03, -2.235681e-03, -2.260750e-03, -2.252383e-03, -2.270724e-03,
            -2.235583e-03, -2.235255e-03, -2.244869e-03, -2.223738e-03, -2.231219e-03,
        ],
        'with_initial_voltage': [
            -1.619374e-03, -1.596250e-03, -1.618585e-03, -1.611130e-03, -1.627472e-03,
            -1.596162e-03, -1.595869e-03, -1.604435e-03, -1.585608e-03, -1.592274e-03,
        ],
        'with_lead_resistance': [
            -2.013923e-03, -1.990812e-03, -2.013134e-03, -2.005684e-03, -2.022016e-03,
            -1.990724e-03, -1.990432e-03, -1.998993e-03, -1.980176e-03, -1.986838e-03,
        ],
        'with_all': [
            -1.818557e-03, -1.792588e-03, -1.817671e-03, -1.809299e-03, -1.827651e-03,
            -1.792490e-03, -1.792161e-03, -1.801781e-03, -1.780638e-03, -1.788123e-03,
        ],
    },
    10188: {
        'baseline': [
            -4.021893e-03, -3.975806e-03, -4.020319e-03, -4.005462e-03, -4.038031e-03,
            -3.975631e-03, -3.975048e-03, -3.992120e-03, -3.954596e-03, -3.967881e-03,
        ],
        'with_adjustment': [
            -4.516585e-03, -4.464830e-03, -4.514819e-03, -4.498134e-03, -4.534709e-03,
            -4.464633e-03, -4.463979e-03, -4.483151e-03, -4.441012e-03, -4.455931e-03,
        ],
        'with_initial_voltage': [
            -3.234898e-03, -3.188758e-03, -3.233323e-03, -3.218449e-03, -3.251055e-03,
            -3.188583e-03, -3.188000e-03, -3.205091e-03, -3.167524e-03, -3.180824e-03,
        ],
        'with_lead_resistance': [
            -4.036073e-03, -3.989823e-03, -4.034494e-03, -4.019585e-03, -4.052268e-03,
            -3.989648e-03, -3.989063e-03, -4.006195e-03, -3.968539e-03, -3.981871e-03,
        ],
        'with_all': [
            -3.645599e-03, -3.593601e-03, -3.643824e-03, -3.627061e-03, -3.663807e-03,
            -3.593403e-03, -3.592746e-03, -3.612008e-03, -3.569671e-03, -3.584660e-03,
        ],
    },
    10189: {
        'baseline': [
            -2.621981e-03, -2.591848e-03, -2.620952e-03, -2.611238e-03, -2.632533e-03,
            -2.591733e-03, -2.591352e-03, -2.602514e-03, -2.577981e-03, -2.586667e-03,
        ],
        'with_adjustment': [
            -2.944485e-03, -2.910645e-03, -2.943330e-03, -2.932420e-03, -2.956335e-03,
            -2.910517e-03, -2.910089e-03, -2.922624e-03, -2.895073e-03, -2.904827e-03,
        ],
        'with_initial_voltage': [
            -2.107695e-03, -2.077562e-03, -2.106667e-03, -2.096952e-03, -2.118248e-03,
            -2.077448e-03, -2.077067e-03, -2.088229e-03, -2.063695e-03, -2.072381e-03,
        ],
        'with_lead_resistance': [
            -2.631225e-03, -2.600986e-03, -2.630193e-03, -2.620445e-03, -2.641815e-03,
            -2.600871e-03, -2.600489e-03, -2.611690e-03, -2.587070e-03, -2.595787e-03,
        ],
        'with_all': [
            -2.375287e-03, -2.341328e-03, -2.374128e-03, -2.363180e-03, -2.387179e-03,
            -2.341199e-03, -2.340770e-03, -2.353349e-03, -2.325701e-03, -2.335489e-03,
        ],
    },
    10271: {
        'baseline': [
            -5.215246e-03, -5.155634e-03, -5.213211e-03, -5.193994e-03, -5.236120e-03,
            -5.155408e-03, -5.154654e-03, -5.176736e-03, -5.128199e-03, -5.145384e-03,
        ],
        'with_adjustment': [
            -5.856721e-03, -5.789777e-03, -5.854436e-03, -5.832856e-03, -5.880162e-03,
            -5.789523e-03, -5.788676e-03, -5.813475e-03, -5.758968e-03, -5.778266e-03,
        ],
        'with_initial_voltage': [
            -4.196815e-03, -4.137074e-03, -4.194776e-03, -4.175517e-03, -4.217733e-03,
            -4.136848e-03, -4.136092e-03, -4.158222e-03, -4.109581e-03, -4.126802e-03,
        ],
        'with_lead_resistance': [
            -5.233633e-03, -5.173811e-03, -5.231592e-03, -5.212307e-03, -5.254581e-03,
            -5.173584e-03, -5.172828e-03, -5.194988e-03, -5.146280e-03, -5.163525e-03,
        ],
        'with_all': [
            -4.729640e-03, -4.662315e-03, -4.727342e-03, -4.705639e-03, -4.753214e-03,
            -4.662059e-03, -4.661208e-03, -4.686147e-03, -4.631330e-03, -4.650738e-03,
        ],
    },
    10272: {
        'baseline': [
            -5.215246e-03, -5.155634e-03, -5.213211e-03, -5.193994e-03, -5.236120e-03,
            -5.155408e-03, -5.154654e-03, -5.176736e-03, -5.128199e-03, -5.145384e-03,
        ],
        'with_adjustment': [
            -5.856721e-03, -5.789777e-03, -5.854436e-03, -5.832856e-03, -5.880162e-03,
            -5.789523e-03, -5.788676e-03, -5.813475e-03, -5.758968e-03, -5.778266e-03,
        ],
        'with_initial_voltage': [
            -4.196815e-03, -4.137074e-03, -4.194776e-03, -4.175517e-03, -4.217733e-03,
            -4.136848e-03, -4.136092e-03, -4.158222e-03, -4.109581e-03, -4.126802e-03,
        ],
        'with_lead_resistance': [
            -5.233633e-03, -5.173811e-03, -5.231592e-03, -5.212307e-03, -5.254581e-03,
            -5.173584e-03, -5.172828e-03, -5.194988e-03, -5.146280e-03, -5.163525e-03,
        ],
        'with_all': [
            -4.729640e-03, -4.662315e-03, -4.727342e-03, -4.705639e-03, -4.753214e-03,
            -4.662059e-03, -4.661208e-03, -4.686147e-03, -4.631330e-03, -4.650738e-03,
        ],
    },
}


@pytest.mark.parametrize("configuration", [
    10183, 10184, 10185, 10188, 10189, 10271, 10272])
@pytest.mark.parametrize("key", [
    "baseline", "with_adjustment", "with_initial_voltage", "with_lead_resistance", "with_all"])
def test_strain_scaling(configuration, key):
    data = StubTdmsData(np.array(_strain_voltage))
    expected_data = np.array(_measured_strain[configuration][key])

    lead_wire_resistance = 0.0
    initial_bridge_voltage = 0.0
    gain_adjustment = 1.0

    if key in ("with_adjustment", "with_all"):
        gain_adjustment = 1.123
    if key in ("with_initial_voltage", "with_all"):
        initial_bridge_voltage = 0.00135
    if key in ("with_lead_resistance", "with_all"):
        lead_wire_resistance = 1.234

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Strain",
        "NI_Scale[0]_Strain_Configuration": configuration,
        "NI_Scale[0]_Strain_Poisson_Ratio": 0.3,
        "NI_Scale[0]_Strain_Gage_Resistance": 350.0,
        "NI_Scale[0]_Strain_Lead_Wire_Resistance": lead_wire_resistance,
        "NI_Scale[0]_Strain_Initial_Bridge_Voltage": initial_bridge_voltage,
        "NI_Scale[0]_Strain_Gage_Factor": 2.1,
        "NI_Scale[0]_Strain_Bridge_Shunt_Calibration_Gain_Adjustment": gain_adjustment,
        "NI_Scale[0]_Strain_Voltage_Excitation": 2.5,
        "NI_Scale[0]_Strain_Input_Source": 0xFFFFFFFF,
    }

    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_data)


def test_unsupported_strain_configuration():
    data = StubTdmsData(np.array([0.0, 0.0, 0.0]))

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Strain",
        "NI_Scale[0]_Strain_Configuration": 12345,
        "NI_Scale[0]_Strain_Poisson_Ratio": 0.3,
        "NI_Scale[0]_Strain_Gage_Resistance": 350.0,
        "NI_Scale[0]_Strain_Lead_Wire_Resistance": 0.0,
        "NI_Scale[0]_Strain_Initial_Bridge_Voltage": 0.0,
        "NI_Scale[0]_Strain_Gage_Factor": 2.1,
        "NI_Scale[0]_Strain_Bridge_Shunt_Calibration_Gain_Adjustment": 1.0,
        "NI_Scale[0]_Strain_Voltage_Excitation": 2.5,
        "NI_Scale[0]_Strain_Input_Source": 0xFFFFFFFF,
    }

    scaling = get_scaling(properties, {}, {})

    with pytest.raises(Exception) as exc_info:
        _ = scaling.scale(data)
    assert str(exc_info.value) == "Strain gauge configuration 12345 is not supported"


def test_multiple_scalings_applied_in_order():
    """Test all scalings applied from multiple scalings
    """

    data = StubTdmsData(np.array([1.0, 2.0, 3.0]))
    expected_scaled_data = np.array([21.0, 27.0, 33.0])

    properties = {
        "NI_Number_Of_Scales": 3,
        "NI_Scaling_Status": "unscaled",
        "NI_Scale[0]_Scale_Type": "Linear",
        "NI_Scale[0]_Linear_Slope": 1.0,
        "NI_Scale[0]_Linear_Y_Intercept": 1.0,
        "NI_Scale[0]_Linear_Input_Source": 0xFFFFFFFF,
        "NI_Scale[1]_Scale_Type": "Linear",
        "NI_Scale[1]_Linear_Slope": 2.0,
        "NI_Scale[1]_Linear_Y_Intercept": 2.0,
        "NI_Scale[1]_Linear_Input_Source": 0,
        "NI_Scale[2]_Scale_Type": "Linear",
        "NI_Scale[2]_Linear_Slope": 3.0,
        "NI_Scale[2]_Linear_Y_Intercept": 3.0,
        "NI_Scale[2]_Linear_Input_Source": 1,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_multiple_scalings_but_all_with_raw_data_input():
    """Test that only the last scaling is applied from multiple scalings
       when it has the raw data as the input source
    """

    data = StubTdmsData(np.array([1.0, 2.0, 3.0]))
    expected_scaled_data = np.array([6.0, 9.0, 12.0])

    properties = {
        "NI_Number_Of_Scales": 3,
        "NI_Scale[0]_Scale_Type": "Linear",
        "NI_Scale[0]_Linear_Slope": 1.0,
        "NI_Scale[0]_Linear_Y_Intercept": 1.0,
        "NI_Scale[0]_Linear_Input_Source": 0xFFFFFFFF,
        "NI_Scale[1]_Scale_Type": "Linear",
        "NI_Scale[1]_Linear_Slope": 2.0,
        "NI_Scale[1]_Linear_Y_Intercept": 2.0,
        "NI_Scale[1]_Linear_Input_Source": 0xFFFFFFFF,
        "NI_Scale[2]_Scale_Type": "Linear",
        "NI_Scale[2]_Linear_Slope": 3.0,
        "NI_Scale[2]_Linear_Y_Intercept": 3.0,
        "NI_Scale[2]_Linear_Input_Source": 0xFFFFFFFF,
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_scaling_from_group():
    """Test linear scaling in a group"""

    data = StubTdmsData(np.array([1.0, 2.0, 3.0]))
    expected_scaled_data = np.array([12.0, 14.0, 16.0])

    group_properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Linear",
        "NI_Scale[0]_Linear_Slope": 2.0,
        "NI_Scale[0]_Linear_Y_Intercept": 10.0,
    }
    scaling = get_scaling({}, group_properties, {})
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_scaling_from_root():
    """Test linear scaling in the root object"""

    data = StubTdmsData(np.array([1.0, 2.0, 3.0]))
    expected_scaled_data = np.array([12.0, 14.0, 16.0])

    root_properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Linear",
        "NI_Scale[0]_Linear_Slope": 2.0,
        "NI_Scale[0]_Linear_Y_Intercept": 10.0,
    }
    scaling = get_scaling({}, {}, root_properties)
    scaled_data = scaling.scale(data)

    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


def test_scaling_status_scaled():
    """ When the scaling status is scaled, data is already scaled so scaling should not be applied
    """
    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "Linear",
        "NI_Scale[0]_Linear_Slope": 2.0,
        "NI_Scale[0]_Linear_Y_Intercept": 10.0,
        "NI_Scaling_Status": "scaled",
    }
    scaling = get_scaling(properties, {}, {})
    assert scaling is None


def test_advanced_api_scaling():
    """Test AdvancedAPI scaling (no-op)"""

    data = StubTdmsData(np.array([1, 2, 3], dtype=np.dtype('int32')))
    expected_scaled_data = np.array([1, 2, 3])

    properties = {
        "NI_Number_Of_Scales": 1,
        "NI_Scale[0]_Scale_Type": "AdvancedAPI",
    }
    scaling = get_scaling(properties, {}, {})
    scaled_data = scaling.scale(data)

    assert scaling.get_dtype(types.Int32, None) == np.dtype('int32')
    assert scaled_data.dtype == np.dtype('int32')
    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)


class StubTdmsData(object):
    def __init__(self, data):
        self.data = data
        self.scaler_data = None


class StubDaqmxData(object):
    def __init__(self, scaler_data):
        self.data = None
        self.scaler_data = scaler_data



from itertools import zip_longest


toc_properties = {
    'kTocMetaData': (1 << 1),
    'kTocRawData': (1 << 3),
    'kTocDAQmxRawData': (1 << 7),
    'kTocInterleavedData': (1 << 5),
    'kTocBigEndian': (1 << 6),
    'kTocNewObjList': (1 << 2)
}


class ObjectPath(object):
    """ Represents the path of an object in a TDMS file

        :ivar group: Group name or None for the root object
        :ivar channel: Channel name or None for the root object or a group objecct
    """
    def __init__(self, *path_components):
        self.group = None
        self.channel = None
        if len(path_components) > 0:
            self.group = path_components[0]
        if len(path_components) > 1:
            self.channel = path_components[1]
        if len(path_components) > 2:
            raise ValueError("Object path may only have up to two components")
        self._path = _components_to_path(self.group, self.channel)

    @property
    def is_root(self):
        return self.group is None

    @property
    def is_group(self):
        return self.group is not None and self.channel is None

    @property
    def is_channel(self):
        return self.channel is not None

    def group_path(self):
        """ For channel paths, returns the path of the channel's group as a string
        """
        return _components_to_path(self.group, None)

    @staticmethod
    def from_string(path_string):
        components = list(_path_components(path_string))
        return ObjectPath(*components)

    def __str__(self):
        """ String representation of the object path
        """
        return self._path


def _path_components(path):
    """ Generator that yields components within an object path
    """
    # Iterate over each character and the next character
    chars = zip_longest(path, path[1:])
    try:
        # Iterate over components
        while True:
            char, next_char = next(chars)
            if char != '/':
                raise ValueError("Invalid path, expected \"/\"")
            elif next_char is not None and next_char != "'":
                raise ValueError("Invalid path, expected \"'\"")
            else:
                # Consume "'" or raise StopIteration if at the end
                next(chars)
            component = []
            # Iterate over characters in component name
            while True:
                char, next_char = next(chars)
                if char == "'" and next_char == "'":
                    component += "'"
                    # Consume second "'"
                    next(chars)
                elif char == "'":
                    yield "".join(component)
                    break
                else:
                    component += char
    except StopIteration:
        return


def _components_to_path(group, channel):
    components = []
    if group is not None:
        components.append(group)
    if channel is not None:
        components.append(channel)
    return ('/' + '/'.join(
        ["'" + c.replace("'", "''") + "'" for c in components]))



from functools import wraps
import logging
import time


def cached_property(func):
    """ Wraps a method on a class to make it a property and caches the result the first time it is evaluated
    """
    attr_name = '_cached_prop_' + func.__name__

    @property
    @wraps(func)
    def get(self):
        try:
            return getattr(self, attr_name)
        except AttributeError:
            value = func(self)
            setattr(self, attr_name, value)
            return value

    return get


class Timer(object):
    """ Context manager for logging the  time taken  for an operation
    """

    def __init__(self, log, description):
        self._enabled = log.isEnabledFor(logging.INFO)
        self._log = log
        self._description = description
        self._start_time = None

    def __enter__(self):
        if not self._enabled:
            return self
        self._start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._enabled:
            return
        end_time = time.perf_counter()

        elapsed_time = (end_time - self._start_time) * 1.0e3
        self._log.info("{0}: Took {1} ms".format(self._description, elapsed_time))



"""Module for reading binary TDMS files produced by LabView"""

from __future__ import absolute_import


# Make version number available
from .version import __version_info__, __version__

# Export public objects
from .tdms import TdmsFile, TdmsGroup, TdmsChannel, DataChunk, GroupDataChunk, ChannelDataChunk
from .writer import TdmsWriter, RootObject, GroupObject, ChannelObject



"""Test reading of TDMS files with DAQmx data
"""

from collections import defaultdict
import logging
import numpy as np
import pytest

from nptdms import TdmsFile
from nptdms import types
from nptdms.log import log_manager
from nptdms.test.scenarios import timestamp_hexlify
from nptdms.test.util import (
    GeneratedFile, hexlify_value, string_hexlify, segment_objects_metadata, hex_properties,
    root_metadata, group_metadata)


def test_single_channel_i16():
    """ Test loading a DAQmx file with a single channel of I16 data
    """

    scaler_metadata = daqmx_scaler_metadata(0, 3, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [2], [scaler_metadata]))
    data = (
        "01 00"
        "02 00"
        "FF FF"
        "FE FF"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["Channel1"].raw_data

    assert data.dtype == np.int16
    np.testing.assert_array_equal(data, [1, 2, -1, -2])


def test_single_channel_u16():
    """ Test loading a DAQmx file with a single channel of U16 data
    """

    scaler_metadata = daqmx_scaler_metadata(0, 2, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [2], [scaler_metadata]))
    data = (
        # Data for segment
        "01 00"
        "02 00"
        "FF FF"
        "FE FF"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["Channel1"].raw_data

    assert data.dtype == np.uint16
    np.testing.assert_array_equal(data, [1, 2, 2**16 - 1, 2**16 - 2])


def test_single_channel_i32():
    """ Test loading a DAQmx file with a single channel of I32 data
    """

    scaler_metadata = daqmx_scaler_metadata(0, 5, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_metadata]))
    data = (
        # Data for segment
        "01 00 00 00"
        "02 00 00 00"
        "FF FF FF FF"
        "FE FF FF FF"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["Channel1"].raw_data

    assert data.dtype == np.int32
    np.testing.assert_array_equal(data, [1, 2, -1, -2])


def test_single_channel_u32():
    """ Test loading a DAQmx file with a single channel of U32 data
    """

    scaler_metadata = daqmx_scaler_metadata(0, 4, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_metadata]))
    data = (
        # Data for segment
        "01 00 00 00"
        "02 00 00 00"
        "FF FF FF FF"
        "FE FF FF FF"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["Channel1"].raw_data

    assert data.dtype == np.uint32
    np.testing.assert_array_equal(data, [1, 2, 2**32 - 1, 2**32 - 2])


def test_two_channel_i16():
    """ Test loading a DAQmx file with two channels of I16 data
    """

    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1]),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2]))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"].raw_data
    assert data_1.dtype == np.int16
    np.testing.assert_array_equal(data_1, [1, 2, 3, 4])

    data_2 = tdms_data["Group"]["Channel2"].raw_data
    assert data_2.dtype == np.int16
    np.testing.assert_array_equal(data_2, [17, 18, 19, 20])


def test_two_channels_without_daqmx_toc_flag():
    """ Test loading a DAQmx file with two channels of DAQmx data but where
        the segment header is missing the DAQmx flag (see issue #226)
    """

    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1]),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2]))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc_non_daqmx(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"].raw_data
    assert data_1.dtype == np.int16
    np.testing.assert_array_equal(data_1, [1, 2, 3, 4])

    data_2 = tdms_data["Group"]["Channel2"].raw_data
    assert data_2.dtype == np.int16
    np.testing.assert_array_equal(data_2, [17, 18, 19, 20])


def test_daqmx_metadata_without_daqmx_raw_data():
    """ Test loading a file that uses DAQmx format metadata but where the data types are not raw data (see issue #226)
    """

    scaler_1 = daqmx_scaler_metadata(0, 0xFFFFFFFF, 0)  # Timestamp data
    scaler_2 = daqmx_scaler_metadata(0, 7, 16)
    scaler_3 = daqmx_scaler_metadata(0, 9, 24)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [32], [scaler_1], data_type=types.TimeStamp.enum_value),
        daqmx_channel_metadata("Channel2", 4, [32], [scaler_2], data_type=types.Int64.enum_value),
        daqmx_channel_metadata("Channel3", 4, [32], [scaler_3], data_type=types.DoubleFloat.enum_value))

    times = [
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
        np.datetime64('2012-08-23T12:00:00.0', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
    ]

    data = (
        timestamp_hexlify(times[0]) +
        "01 00 00 00 00 00 00 00" +
        hexlify_value("<d", 1.0) +
        timestamp_hexlify(times[1]) +
        "02 00 00 00 00 00 00 00" +
        hexlify_value("<d", 2.0) +
        timestamp_hexlify(times[2]) +
        "03 00 00 00 00 00 00 00" +
        hexlify_value("<d", 3.0) +
        timestamp_hexlify(times[3]) +
        "04 00 00 00 00 00 00 00" +
        hexlify_value("<d", 4.0)
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc_non_daqmx(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"][:]
    np.testing.assert_array_equal(data_1, times)

    data_2 = tdms_data["Group"]["Channel2"][:]
    assert data_2.dtype == np.int64
    np.testing.assert_array_equal(data_2, [1, 2, 3, 4])

    data_3 = tdms_data["Group"]["Channel3"][:]
    assert data_3.dtype == np.float64
    np.testing.assert_array_equal(data_3, [1.0, 2.0, 3.0, 4.0])


def test_exception_on_mismatch_of_types_for_non_raw_daqmx():
    scaler = daqmx_scaler_metadata(0, 3, 0, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 2, [2], [scaler], data_type=types.Uint16.enum_value))
    data = "01 00 02 00"

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc_non_daqmx(), metadata, data)

    with pytest.raises(ValueError) as exception:
        _ = test_file.load()
    error_message = str(exception.value)

    assert error_message == "Expected scaler data type to be Uint16 but got Int16"


def test_exception_on_multiple_scalers_for_non_raw_daqmx():
    scaler_1 = daqmx_scaler_metadata(0, 3, 0, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 2, [4], [scaler_1, scaler_2], data_type=types.Int32.enum_value))
    data = "01 00 00 00"

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc_non_daqmx(), metadata, data)

    with pytest.raises(ValueError) as exception:
        _ = test_file.load()
    error_message = str(exception.value)

    assert error_message == "Expected only one scaler for channel with type Int32"


def test_mixed_channel_widths():
    """ Test loading a DAQmx file with channels with different widths
    """

    scaler_1 = daqmx_scaler_metadata(0, 1, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 1)
    scaler_3 = daqmx_scaler_metadata(0, 5, 3)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [7], [scaler_1]),
        daqmx_channel_metadata("Channel2", 4, [7], [scaler_2]),
        daqmx_channel_metadata("Channel3", 4, [7], [scaler_3]))
    data = (
        # Data for segment
        "01 11 00 21 00 00 00"
        "02 12 00 22 00 00 00"
        "03 13 00 23 00 00 00"
        "04 14 00 24 00 00 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"].raw_data
    assert data_1.dtype == np.int8
    np.testing.assert_array_equal(data_1, [1, 2, 3, 4])

    data_2 = tdms_data["Group"]["Channel2"].raw_data
    assert data_2.dtype == np.int16
    np.testing.assert_array_equal(data_2, [17, 18, 19, 20])

    data_3 = tdms_data["Group"]["Channel3"].raw_data
    assert data_3.dtype == np.int32
    np.testing.assert_array_equal(data_3, [33, 34, 35, 36])


def test_multiple_scalers_with_same_type():
    """ Test loading a DAQmx file with one channel containing multiple
        format changing scalers of the same type
    """

    scaler_metadata = [
        daqmx_scaler_metadata(0, 3, 0),
        daqmx_scaler_metadata(1, 3, 2)]
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], scaler_metadata))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()
    channel = tdms_data["Group"]["Channel1"]

    scaler_0_data = channel.raw_scaler_data[0]
    assert scaler_0_data.dtype == np.int16
    np.testing.assert_array_equal(scaler_0_data, [1, 2, 3, 4])

    scaler_1_data = channel.raw_scaler_data[1]
    assert scaler_1_data.dtype == np.int16
    np.testing.assert_array_equal(scaler_1_data, [17, 18, 19, 20])


def test_multiple_scalers_with_different_types():
    """ Test loading a DAQmx file with one channel containing multiple
        format changing scalers of different types
    """

    scaler_metadata = [
        daqmx_scaler_metadata(0, 1, 0),
        daqmx_scaler_metadata(1, 3, 1),
        daqmx_scaler_metadata(2, 5, 3)]
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [7], scaler_metadata))
    data = (
        # Data for segment
        "01 11 00 21 00 00 00"
        "02 12 00 22 00 00 00"
        "03 13 00 23 00 00 00"
        "04 14 00 24 00 00 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()
    channel = tdms_data["Group"]["Channel1"]

    scaler_0_data = channel.raw_scaler_data[0]
    assert scaler_0_data.dtype == np.int8
    np.testing.assert_array_equal(scaler_0_data, [1, 2, 3, 4])

    scaler_1_data = channel.raw_scaler_data[1]
    assert scaler_1_data.dtype == np.int16
    np.testing.assert_array_equal(scaler_1_data, [17, 18, 19, 20])

    scaler_2_data = channel.raw_scaler_data[2]
    assert scaler_2_data.dtype == np.int32
    np.testing.assert_array_equal(scaler_2_data, [33, 34, 35, 36])


def test_multiple_raw_data_buffers():
    """ Test loading a DAQmx file with multiple raw data buffers
    """

    scaler_1 = daqmx_scaler_metadata(0, 3, 0, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2, 0)
    scaler_3 = daqmx_scaler_metadata(0, 3, 0, 1)
    scaler_4 = daqmx_scaler_metadata(0, 3, 2, 1)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4, 4], [scaler_1]),
        daqmx_channel_metadata("Channel2", 4, [4, 4], [scaler_2]),
        daqmx_channel_metadata("Channel3", 4, [4, 4], [scaler_3]),
        daqmx_channel_metadata("Channel4", 4, [4, 4], [scaler_4]))
    data = (
        "01 00" "02 00" "03 00" "04 00"
        "05 00" "06 00" "07 00" "08 00"
        "09 00" "0A 00" "0B 00" "0C 00"
        "0D 00" "0E 00" "0F 00" "10 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"].raw_data
    data_2 = tdms_data["Group"]["Channel2"].raw_data
    data_3 = tdms_data["Group"]["Channel3"].raw_data
    data_4 = tdms_data["Group"]["Channel4"].raw_data

    for data in [data_1, data_2, data_3, data_4]:
        assert data.dtype == np.int16

    np.testing.assert_array_equal(data_1, [1, 3, 5, 7])
    np.testing.assert_array_equal(data_2, [2, 4, 6, 8])
    np.testing.assert_array_equal(data_3, [9, 11, 13, 15])
    np.testing.assert_array_equal(data_4, [10, 12, 14, 16])


def test_multiple_raw_data_buffers_with_different_widths():
    """ DAQmx with raw data buffers with different widths
    """

    scaler_1 = daqmx_scaler_metadata(0, 3, 0, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2, 0)
    scaler_3 = daqmx_scaler_metadata(0, 3, 4, 0)
    scaler_4 = daqmx_scaler_metadata(0, 5, 0, 1)
    scaler_5 = daqmx_scaler_metadata(0, 5, 4, 1)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [6, 8], [scaler_1]),
        daqmx_channel_metadata("Channel2", 4, [6, 8], [scaler_2]),
        daqmx_channel_metadata("Channel3", 4, [6, 8], [scaler_3]),
        daqmx_channel_metadata("Channel4", 4, [6, 8], [scaler_4]),
        daqmx_channel_metadata("Channel5", 4, [6, 8], [scaler_5]))
    data = (
        "01 00" "02 00" "03 00"
        "04 00" "05 00" "06 00"
        "07 00" "08 00" "09 00"
        "0A 00" "0B 00" "0C 00"
        "0D 00 00 00" "0E 00 00 00"
        "0F 00 00 00" "10 00 00 00"
        "11 00 00 00" "12 00 00 00"
        "13 00 00 00" "14 00 00 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"].raw_data
    data_2 = tdms_data["Group"]["Channel2"].raw_data
    data_3 = tdms_data["Group"]["Channel3"].raw_data
    data_4 = tdms_data["Group"]["Channel4"].raw_data
    data_5 = tdms_data["Group"]["Channel5"].raw_data

    for data in [data_1, data_2, data_3]:
        assert data.dtype == np.int16
    for data in [data_4, data_5]:
        assert data.dtype == np.int32

    np.testing.assert_array_equal(data_1, [1, 4, 7, 10])
    np.testing.assert_array_equal(data_2, [2, 5, 8, 11])
    np.testing.assert_array_equal(data_3, [3, 6, 9, 12])
    np.testing.assert_array_equal(data_4, [13, 15, 17, 19])
    np.testing.assert_array_equal(data_5, [14, 16, 18, 20])


def test_multiple_raw_data_buffers_with_different_lengths():
    """ DAQmx with raw data buffers with different lengths
    """

    properties = {
        "NI_Number_Of_Scales": (3, "01 00 00 00"),
    }
    scaler_1 = daqmx_scaler_metadata(0, 3, 0, 0)
    scaler_2 = daqmx_scaler_metadata(0, 5, 2, 0)
    scaler_3 = daqmx_scaler_metadata(0, 3, 0, 1)
    scaler_4 = daqmx_scaler_metadata(0, 5, 2, 1)
    scaler_5 = daqmx_scaler_metadata(0, 3, 0, 2)
    scaler_6 = daqmx_scaler_metadata(0, 5, 2, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [6, 6, 6], [scaler_1], properties=properties),
        daqmx_channel_metadata("Channel2", 4, [6, 6, 6], [scaler_2], properties=properties),
        daqmx_channel_metadata("Channel3", 2, [6, 6, 6], [scaler_3], properties=properties),
        daqmx_channel_metadata("Channel4", 2, [6, 6, 6], [scaler_4], properties=properties),
        daqmx_channel_metadata("Channel5", 1, [6, 6, 6], [scaler_5], properties=properties),
        daqmx_channel_metadata("Channel6", 1, [6, 6, 6], [scaler_6], properties=properties))
    data = (
        # Chunk 1
        # Buffer 0
        "01 00" "02 00 00 00"
        "01 00" "02 00 00 00"
        "01 00" "02 00 00 00"
        "01 00" "02 00 00 00"
        # Buffer 1
        "03 00" "04 00 00 00"
        "03 00" "04 00 00 00"
        # Buffer 2
        "05 00" "06 00 00 00"
        # Chunk 2
        # Buffer 0
        "01 00" "02 00 00 00"
        "01 00" "02 00 00 00"
        "01 00" "02 00 00 00"
        "01 00" "02 00 00 00"
        # Buffer 1
        "03 00" "04 00 00 00"
        "03 00" "04 00 00 00"
        # Buffer 2
        "05 00" "06 00 00 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data_1 = tdms_data["Group"]["Channel1"][:]
    data_2 = tdms_data["Group"]["Channel2"][:]
    data_3 = tdms_data["Group"]["Channel3"][:]
    data_4 = tdms_data["Group"]["Channel4"][:]
    data_5 = tdms_data["Group"]["Channel5"][:]
    data_6 = tdms_data["Group"]["Channel6"][:]

    for data in [data_1, data_3, data_5]:
        assert data.dtype == np.int16
    for data in [data_2, data_4, data_6]:
        assert data.dtype == np.int32

    np.testing.assert_array_equal(data_1, [1] * 8)
    np.testing.assert_array_equal(data_2, [2] * 8)
    np.testing.assert_array_equal(data_3, [3] * 4)
    np.testing.assert_array_equal(data_4, [4] * 4)
    np.testing.assert_array_equal(data_5, [5] * 2)
    np.testing.assert_array_equal(data_6, [6] * 2)


def test_incomplete_segment_with_different_length_buffers():
    """ DAQmx with raw data buffers with different lengths
    """

    scaler_1 = daqmx_scaler_metadata(0xFFFFFFFF, 2, 0, 0)
    scaler_2 = daqmx_scaler_metadata(0xFFFFFFFF, 2, 2, 0)
    scaler_3 = daqmx_scaler_metadata(0xFFFFFFFF, 2, 0, 1)
    scaler_4 = daqmx_scaler_metadata(0xFFFFFFFF, 2, 2, 1)
    scaler_5 = daqmx_scaler_metadata(0xFFFFFFFF, 2, 0, 2)
    scaler_6 = daqmx_scaler_metadata(0xFFFFFFFF, 2, 2, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4, 4, 4], [scaler_1], data_type=types.Uint16.enum_value),
        daqmx_channel_metadata("Channel2", 4, [4, 4, 4], [scaler_2], data_type=types.Uint16.enum_value),
        daqmx_channel_metadata("Channel3", 2, [4, 4, 4], [scaler_3], data_type=types.Uint16.enum_value),
        daqmx_channel_metadata("Channel4", 2, [4, 4, 4], [scaler_4], data_type=types.Uint16.enum_value),
        daqmx_channel_metadata("Channel5", 1, [4, 4, 4], [scaler_5], data_type=types.Uint16.enum_value),
        daqmx_channel_metadata("Channel6", 1, [4, 4, 4], [scaler_6], data_type=types.Uint16.enum_value))
    data = (
        # Chunk 1
        # Buffer 0
        "01 00" "02 00"
        "01 00" "02 00"
        "01 00" "02 00"
        "01 00" "02 00"
        # Buffer 1
        "03 00" "04 00"
        "03 00" "04 00"
        # Buffer 2
        "05 00" "06 00"
        # Chunk 2
        # Buffer 0
        "01 00" "02 00"
        "01 00" "02 00"
        "01 00" "02 00"
        "01 00" "02 00"
        # Buffer 1
        "03 00" "04 00"
        "03 00" "04 00"
        # Buffer 2
        "05 00" "06 00"
        # Incomplete third chunk
        # Buffer 0
        "01 00" "02 00"
        "01 00" "02 00"
        "01 00" "02 00"
        "01 00" "02 00"
        # Buffer 1
        "03 00" "04 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc_non_daqmx(), metadata, data, incomplete=True)
    tdms_data = test_file.load()

    group = tdms_data["Group"]

    np.testing.assert_array_equal(group["Channel1"][:], [1] * 12)
    np.testing.assert_array_equal(group["Channel2"][:], [2] * 12)
    np.testing.assert_array_equal(group["Channel3"][:], [3] * 5)
    np.testing.assert_array_equal(group["Channel4"][:], [4] * 5)
    np.testing.assert_array_equal(group["Channel5"][:], [5] * 2)
    np.testing.assert_array_equal(group["Channel6"][:], [6] * 2)


def test_multiple_raw_data_buffers_with_scalers_split_across_buffers():
    """ DAQmx with scalers split across different raw data buffers
    """

    scaler_1 = daqmx_scaler_metadata(0, 3, 0, 0)
    scaler_2 = daqmx_scaler_metadata(1, 3, 0, 1)
    scaler_3 = daqmx_scaler_metadata(0, 3, 2, 0)
    scaler_4 = daqmx_scaler_metadata(1, 3, 2, 1)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata(
            "Channel1", 4, [4, 4], [scaler_1, scaler_2]),
        daqmx_channel_metadata(
            "Channel2", 4, [4, 4], [scaler_3, scaler_4]))
    data = (
        "01 00" "02 00" "03 00" "04 00"
        "05 00" "06 00" "07 00" "08 00"
        "09 00" "0A 00" "0B 00" "0C 00"
        "0D 00" "0E 00" "0F 00" "10 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    channel_1 = tdms_data["Group"]["Channel1"]
    channel_2 = tdms_data["Group"]["Channel2"]

    scaler_data_1 = channel_1.raw_scaler_data[0]
    scaler_data_2 = channel_1.raw_scaler_data[1]
    scaler_data_3 = channel_2.raw_scaler_data[0]
    scaler_data_4 = channel_2.raw_scaler_data[1]

    for data in [
            scaler_data_1, scaler_data_2, scaler_data_3, scaler_data_4]:
        assert data.dtype == np.int16

    np.testing.assert_array_equal(scaler_data_1, [1, 3, 5, 7])
    np.testing.assert_array_equal(scaler_data_2, [9, 11, 13, 15])
    np.testing.assert_array_equal(scaler_data_3, [2, 4, 6, 8])
    np.testing.assert_array_equal(scaler_data_4, [10, 12, 14, 16])


def test_digital_line_scaler_data():
    """ Test loading a DAQmx file with a single channel of U8 digital line scaler data
    """

    scaler_metadata = digital_scaler_metadata(0, 0, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_metadata], digital_line_scaler=True))
    data = (
        "00 00 00 00"
        "01 00 00 00"
        "00 00 00 00"
        "01 00 00 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["Channel1"].raw_data

    assert data.dtype == np.uint8
    np.testing.assert_array_equal(data, [0, 1, 0, 1])


@pytest.mark.parametrize('byte_offset', [0, 1, 2, 3])
def test_digital_line_scaler_with_multiple_channels(byte_offset):
    """ Test DAQmx digital line scaler data with multiple channels
    """

    scaler_metadata_0 = digital_scaler_metadata(0, 0, byte_offset * 8 + 0)
    scaler_metadata_1 = digital_scaler_metadata(0, 0, byte_offset * 8 + 1)
    scaler_metadata_2 = digital_scaler_metadata(0, 0, byte_offset * 8 + 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel0", 4, [4], [scaler_metadata_0], digital_line_scaler=True),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_metadata_1], digital_line_scaler=True),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_metadata_2], digital_line_scaler=True),
    )
    byte_values = [
        "00",
        "01",
        "02",
        "03",
        "04",
        "05",
        "06",
        "07",
    ]
    hex_data = " ".join("00" * byte_offset + b + "00" * (3 - byte_offset) for b in byte_values)

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, hex_data)
    tdms_data = test_file.load()

    for (channel_name, expected_data) in [
        ("Channel0", [0, 1, 0, 1, 0, 1, 0, 1]),
        ("Channel1", [0, 0, 1, 1, 0, 0, 1, 1]),
        ("Channel2", [0, 0, 0, 0, 1, 1, 1, 1]),
    ]:
        data = tdms_data["Group"][channel_name].raw_data

        assert data.dtype == np.uint8
        np.testing.assert_array_equal(data, expected_data, "Incorrect data for channel '%s'" % channel_name)


def test_exception_on_mismatch_of_raw_data_widths():
    scaler_1 = daqmx_scaler_metadata(0, 3, 0, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 1, [2], [scaler_1]),
        daqmx_channel_metadata("Channel2", 1, [4], [scaler_2]))
    data = "01 00 02 00"

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)

    with pytest.raises(ValueError) as exception:
        _ = test_file.load()
    error_message = str(exception.value)

    assert (error_message == "Raw data widths for object DaqmxSegmentObject(/'Group'/'Channel2') ([4]) "
            "do not match previous widths ([2])")


def test_lazily_reading_channel():
    """ Test loading channels individually from a DAQmx file
    """

    # Single scale which is just the raw DAQmx scaler data
    properties = {
        "NI_Number_Of_Scales": (3, "01 00 00 00"),
    }
    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1], properties),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2], properties))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            data_1 = tdms_file["Group"]["Channel1"].read_data()
            assert data_1.dtype == np.int16
            np.testing.assert_array_equal(data_1, [1, 2, 3, 4])

            data_2 = tdms_file["Group"]["Channel2"].read_data()
            assert data_2.dtype == np.int16
            np.testing.assert_array_equal(data_2, [17, 18, 19, 20])


def test_lazily_reading_a_subset_of_channel_data():
    """ Test loading a subset of channel data from a DAQmx file
    """

    # Single scale which is just the raw DAQmx scaler data
    properties = {
        "NI_Number_Of_Scales": (3, "01 00 00 00"),
    }
    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1], properties),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2], properties))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            data_1 = tdms_file["Group"]["Channel1"].read_data(1, 2)
            assert data_1.dtype == np.int16
            np.testing.assert_array_equal(data_1, [2, 3])

            data_2 = tdms_file["Group"]["Channel2"].read_data(1, 2)
            assert data_2.dtype == np.int16
            np.testing.assert_array_equal(data_2, [18, 19])


def test_lazily_reading_a_subset_of_raw_channel_data():
    """ Test loading a subset of raw scaler channel data from a DAQmx file
    """

    # Single scale which is just the raw DAQmx scaler data
    properties = {
        "NI_Number_Of_Scales": (3, "01 00 00 00"),
    }
    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1], properties),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2], properties))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            data_1 = tdms_file["Group"]["Channel1"].read_data(1, 2, scaled=False)
            assert len(data_1) == 1
            assert data_1[0].dtype == np.int16
            np.testing.assert_array_equal(data_1[0], [2, 3])

            data_2 = tdms_file["Group"]["Channel2"].read_data(1, 2, scaled=False)
            assert len(data_2) == 1
            assert data_2[0].dtype == np.int16
            np.testing.assert_array_equal(data_2[0], [18, 19])


@pytest.mark.parametrize('offset,length', [
    (0, None),
    (1, None),
    (0, 2),
    (1, 2),
])
def test_read_raw_data(offset, length):
    # Single scale which is just the raw DAQmx scaler data
    properties = {
        "NI_Number_Of_Scales": (3, "01 00 00 00"),
    }
    scaler_1 = daqmx_scaler_metadata(1, 3, 0)
    scaler_2 = daqmx_scaler_metadata(2, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1], properties),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2], properties))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)

    end = None if length is None else offset + length
    with test_file.get_tempfile() as temp_file:
        tdms_file = TdmsFile.read(temp_file.file)
        data_1 = tdms_file["Group"]["Channel1"].read_data(offset=offset, length=length, scaled=False)
        assert data_1[1].dtype == np.int16
        np.testing.assert_array_equal(data_1[1], [1, 2, 3, 4][offset:end])

        data_2 = tdms_file["Group"]["Channel2"].read_data(offset=offset, length=length, scaled=False)
        assert data_2[2].dtype == np.int16
        np.testing.assert_array_equal(data_2[2], [17, 18, 19, 20][offset:end])


def test_stream_data_chunks():
    """Test streaming chunks of DAQmx data from a TDMS file
    """
    properties = {
        "NI_Number_Of_Scales": (3, "01 00 00 00"),
    }
    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1], properties),
        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2], properties))
    data = (
        # Data for segment
        "01 00" "11 00"
        "02 00" "12 00"
        "03 00" "13 00"
        "04 00" "14 00"
        "05 00" "15 00"
        "06 00" "16 00"
        "07 00" "17 00"
        "08 00" "18 00"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)
    data_arrays = defaultdict(list)
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for chunk in tdms_file.data_chunks():
                for group in chunk.groups():
                    for channel in group.channels():
                        key = (group.name, channel.name)
                        assert channel.offset == len(data_arrays[key])
                        data_arrays[key].extend(channel[:])

    expected_channel_data = {
        ("Group", "Channel1"): [1, 2, 3, 4, 5, 6, 7, 8],
        ("Group", "Channel2"): [17, 18, 19, 20, 21, 22, 23, 24],
    }
    for ((group, channel), expected_data) in expected_channel_data.items():
        actual_data = data_arrays[(group, channel)]
        np.testing.assert_equal(actual_data, expected_data)


def test_daqmx_debug_logging(caplog):
    """ Test loading a DAQmx file with debug logging enabled
    """
    scaler_metadata = daqmx_scaler_metadata(0, 3, 0)
    metadata = segment_objects_metadata(
        root_metadata(),
        group_metadata(),
        daqmx_channel_metadata("Channel1", 4, [2], [scaler_metadata]))
    data = (
        "01 00"
        "02 00"
        "FF FF"
        "FE FF"
    )

    test_file = GeneratedFile()
    test_file.add_segment(segment_toc(), metadata, data)

    log_manager.set_level(logging.DEBUG)
    _ = test_file.load()

    assert "Reading metadata for object /'Group'/'Channel1' with index header 0x00001269" in caplog.text
    assert "scale_id=0" in caplog.text
    assert "data_type=Int16" in caplog.text


def segment_toc():
    return (
        "kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocDAQmxRawData")


def segment_toc_non_daqmx():
    return (
        "kTocMetaData", "kTocRawData", "kTocNewObjList")


def daqmx_scaler_metadata(scale_id, type_id, byte_offset, raw_buffer_index=0):
    return (
        # DAQmx data type (type ids don't match TDMS types)
        hexlify_value("<I", type_id) +
        # Raw buffer index
        hexlify_value("<I", raw_buffer_index) +
        # Raw byte offset
        hexlify_value("<I", byte_offset) +
        # Sample format bitmap (don't know what this is for...)
        "00 00 00 00" +
        # Scale ID
        hexlify_value("<I", scale_id))


def digital_scaler_metadata(scale_id, type_id, bit_offset, raw_buffer_index=0):
    return (
        # DAQmx data type (type ids don't match TDMS types)
        hexlify_value("<I", type_id) +
        # Raw buffer index
        hexlify_value("<I", raw_buffer_index) +
        # Raw byte offset
        hexlify_value("<I", bit_offset) +
        # Sample format bitmap (don't know what this is for...)
        "00" +
        # Scale ID
        hexlify_value("<I", scale_id))


def daqmx_channel_metadata(
        channel_name, num_values,
        raw_data_widths, scaler_metadata, properties=None, digital_line_scaler=False,
        data_type=None):
    path = "/'Group'/'" + channel_name + "'"
    # Default to DAQmx data type
    data_type = 0xFFFFFFFF if data_type is None else data_type
    return (
        # Length of the object path
        hexlify_value("<I", len(path)) +
        # Object path
        string_hexlify(path) +
        # Raw data index (DAQmx)
        ("6A 12 00 00" if digital_line_scaler else "69 12 00 00") +
        # Data type
        hexlify_value("<I", data_type) +
        # Array  dimension
        "01 00 00 00" +
        # Number of values (chunk size)
        hexlify_value("<Q", num_values) +
        # Scaler metadata
        hexlify_value("<I", len(scaler_metadata)) +
        "".join(scaler_metadata) +
        # Raw data width vector size
        hexlify_value("<I", len(raw_data_widths)) +
        # Raw data width values
        "".join(hexlify_value("<I", v) for v in raw_data_widths) +
        hex_properties(properties))



"""Module for writing TDMS files"""

from collections import OrderedDict
from datetime import datetime
from io import UnsupportedOperation, BytesIO
import os

import numpy as np
from nptdms.common import toc_properties, ObjectPath
from nptdms.types import *
from nptdms import TdmsFile


class TdmsWriter(object):
    """Writes to a TDMS file.

    A TdmsWriter should be used as a context manager, for example::

        with TdmsWriter(path) as tdms_writer:
            tdms_writer.write_segment(segment_data)
    """

    @classmethod
    def defragment(cls, source, destination, version=4712, index_file=False):
        """ Defragemnts an existing TdmsFile by loading and moving each Object to a separate channel
        to stream read one consecutive part of the file for faster access.

        :param source: Either the path to the tdms file to read
            as a string or pathlib.Path, or an already opened file.
        :param destination: Either the path to the tdms file
        :param version: The TDMS format version to write, which must be either 4712 (the default) or 4713.
            It's important that if you are appending segments to an
            existing TDMS file, this matches the existing file version (this can be queried with the
            :py:attr:`~nptdms.TdmsFile.tdms_version` property).
        :param index_file: Depends on the ``destination`` input.
            If ``destination`` is a path ``index_file`` can either be ``True`` or ``False`` to store a ``.tdms_index``
            file at the same folder location or not.
            If ``destination`` is a readable object ``index_file`` can either be a redable object or ``False``
            to store a ``.tdms_index`` file inside of the submitted object or not.
        """
        file = TdmsFile(source)
        with cls(destination, version=version, index_file=index_file) as new_file:
            new_file.write_segment([RootObject(file.properties)])
            for group in file.groups():
                new_file.write_segment([GroupObject(group.name, group.properties)])
                for channel in group.channels():
                    new_file.write_segment([ChannelObject(
                        group.name,
                        channel.name,
                        channel.read_data(scaled=False),
                        channel.properties
                    )])

    def __init__(self, file, mode='w', version=4712, index_file=False):
        """Initialise a new TDMS writer

        :param file: Either the path to the tdms file, an already
            opened file or a bytes stream.
        :param mode: The mode to open the file with, used when ``file`` is a file path.
            This will be passed through to Python's ``open`` function with 'b' appended
            to ensure the file is opened in binary mode.
            For example, use 'w' (the default) to open a new file or 'a' to append to an existing TDMS file.
        :param version: The TDMS format version to write, which must be either 4712 (the default) or 4713.
            It's important that if you are appending segments to an
            existing TDMS file, this matches the existing file version (this can be queried with the
            :py:attr:`~nptdms.TdmsFile.tdms_version` property).
        :param index_file: Whether or not to write a index file besides the data file. Index files
            can be used to accelerate reading speeds for faster channel extraction and data positions inside
            the data files. If ``file```variable is a path ``index_file`` can be ``True`` to store a ``.tdms_index``
            file at the same folder location or ``False`` to only write the data ``.tdms`` file. If ``file`` variable
            is a readable object ``index_file`` can either be a readable object to write into or ``False`` to omit.
        """
        valid_versions = (4712, 4713)
        if version not in valid_versions:
            raise ValueError("version must be one of %s" % ",".join("%d" % v for v in valid_versions))
        self._file = None
        self._index_file = None
        self._file_path = None
        self._index_file_path = None
        self._file_mode = mode
        self._tdms_version = version

        if hasattr(file, "read"):
            # Is a file
            self._file = file
            if hasattr(index_file, "read"):
                self._index_file = index_file
            elif isinstance(index_file, bool) and not index_file:
                pass
            else:
                raise ValueError(
                    f"Invalid type, ``index_file`` can only be ``False`` or a stream to write into, "
                    "but is {type(index_file)}"
                )
        else:
            self._file_path = file
            if isinstance(index_file, bool):
                if index_file:
                    self._index_file_path = file + "_index"
            else:
                raise ValueError(
                    f"Invalid type, ``index_file`` can  only be ``False`` or ``True`` but is {type(index_file)}."
                )

    def open(self):
        if self._file_path is not None:
            self._file = open(self._file_path, self._file_mode + 'b')
            if self._index_file_path is not None:
                self._index_file = open(self._index_file_path, self._file_mode + 'b')

    def close(self):
        if self._file_path is not None:
            self._file.close()

        if self._index_file_path is not None:
            self._index_file.close()

        self._file = None
        self._index_file = None

    def write_segment(self, objects):
        """ Write a segment of data to a TDMS file

        :param objects: A list of TdmsObject instances to write
        """
        segment = TdmsSegment(objects, version=self._tdms_version)
        segment.write(self._file)

        if self._index_file is not None:
            segment = TdmsSegment(objects, is_index_file=True, version=self._tdms_version)
            segment.write(self._index_file)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


class TdmsSegment(object):
    """A segment of data to be written to a file
    """

    def __init__(self, objects, is_index_file=False, version=4712):
        """Initialise a new segment of TDMS data

        :param objects: A list of TdmsObject instances.
        :param is_index_file: Whether a written file is a data file (.tdms) or a index file (.tdms_index).
        :param version: The TDMS format version to write, which must be either 4712 (the default) or 4713.
        """
        paths = set(obj.path for obj in objects)
        if len(paths) != len(objects):
            raise ValueError("Duplicate object paths found")

        self.objects = objects
        self._tdms_version = version
        self.is_index_file = is_index_file

    def write(self, file):
        metadata = self.metadata()
        metadata_size = sum(len(val.bytes) for val in metadata)

        toc = ['kTocMetaData', 'kTocRawData', 'kTocNewObjList']
        leadin = self.leadin(toc, metadata_size)

        file.write(b''.join(val.bytes for val in leadin))
        file.write(b''.join(val.bytes for val in metadata))
        if not self.is_index_file:
            self._write_data(file)

    def metadata(self):
        metadata = []
        metadata.append(Uint32(len(self.objects)))
        for obj in self.objects:
            metadata.append(String(obj.path))
            metadata.extend(self.raw_data_index(obj))
            properties = read_properties_dict(obj.properties)
            num_properties = len(properties)
            metadata.append(Uint32(num_properties))
            for prop_name, prop_value in properties.items():
                metadata.append(String(prop_name))
                metadata.append(Int32(prop_value.enum_value))
                metadata.append(prop_value)
        return metadata

    def raw_data_index(self, obj):
        if hasattr(obj, 'data'):
            data_type = Int32(obj.data_type.enum_value)
            dimension = Uint32(1)
            num_values = Uint64(len(obj.data))

            data_index = [Uint32(20), data_type, dimension, num_values]
            # For strings, we also need to write the total data size in bytes
            if obj.data_type == String:
                total_size = object_data_size(obj.data_type, obj.data)
                data_index.append(Uint64(total_size))

            return data_index
        else:
            return [Bytes(b'\xFF\xFF\xFF\xFF')]

    def leadin(self, toc, metadata_size):
        leadin = []
        leadin.append(Bytes(b'TDSh' if self.is_index_file else b'TDSm'))

        toc_mask = 0
        for toc_flag in toc:
            toc_mask = toc_mask | toc_properties[toc_flag]
        leadin.append(Int32(toc_mask))

        leadin.append(Int32(self._tdms_version))

        next_segment_offset = metadata_size + self._data_size()
        raw_data_offset = metadata_size
        leadin.append(Uint64(next_segment_offset))
        leadin.append(Uint64(raw_data_offset))

        return leadin

    def _data_size(self):
        data_size = 0
        for obj in self.objects:
            if hasattr(obj, 'data'):
                data_size += object_data_size(obj.data_type, obj.data)
        return data_size

    def _write_data(self, file):
        for obj in self.objects:
            if hasattr(obj, 'data'):
                write_data(file, obj)


class TdmsObject(object):
    @property
    def has_data(self):
        return False

    @property
    def data_type(self):
        return None

    @property
    def path(self):
        return None


class RootObject(TdmsObject):
    """The root TDMS object containing properties for the TDMS file
    """
    def __init__(self, properties=None):
        """Initialise a new GroupObject

        :param properties: A dictionary mapping property names to
            their value.
        """
        self.properties = properties

    @property
    def path(self):
        """The string representation of the root path
        """
        return "/"


class GroupObject(TdmsObject):
    """A TDMS object for a group
    """

    def __init__(self, group, properties=None):
        """Initialise a new GroupObject

        :param group: The name of this group.
        :param properties: A dictionary mapping property names to
            their value.
        """
        self.group = group
        self.properties = properties

    @property
    def path(self):
        """The string representation of this group's path
        """
        return str(ObjectPath(self.group))


class ChannelObject(TdmsObject):
    """A TDMS object for a channel with data
    """

    def __init__(self, group, channel, data, properties=None):
        """Initialise a new ChannelObject

        :param group: The name of the group this channel is in.
        :param channel: The name of this channel.
        :param data: 1-D Numpy array of data to be written.
        :param properties: A dictionary mapping property names to
            their value.
        """
        self.group = group
        self.channel = channel
        self.data = _to_np_array(data)
        if self.data.ndim != 1:
            raise ValueError("Channel data must be a 1d array")
        self.properties = properties

    @property
    def has_data(self):
        return True

    @property
    def data_type(self):
        try:
            return numpy_data_types[self.data.dtype]
        except (AttributeError, KeyError):
            try:
                return _to_tdms_value(self.data[0]).__class__
            except IndexError:
                return Void

    @property
    def path(self):
        """The string representation of this channel's path
        """
        return str(ObjectPath(self.group, self.channel))


def read_properties_dict(properties_dict):
    if properties_dict is None:
        return {}

    return OrderedDict(
        (key, _to_tdms_value(val))
        for key, val in properties_dict.items())


def _to_tdms_value(value):
    if isinstance(value, np.number):
        return numpy_data_types[value.dtype](value)
    if isinstance(value, TdmsType):
        return value
    if isinstance(value, bool) or isinstance(value, np.bool_):
        return Boolean(value)
    if isinstance(value, int):
        return to_int_property_value(value)
    if isinstance(value, float):
        return DoubleFloat(value)
    if isinstance(value, datetime):
        return TimeStamp(value)
    if isinstance(value, np.datetime64):
        return TimeStamp(value)
    if isinstance(value, str):
        return String(value)
    if isinstance(value, bytes):
        return String(value)
    raise TypeError("Unsupported property type for %r" % value)


def to_int_property_value(value):
    if value >= 2 ** 63:
        return Uint64(value)
    if value >= 2 ** 31 or value < -2 ** 31:
        return Int64(value)
    return Int32(value)


def write_data(file, tdms_object):
    if tdms_object.data_type == TimeStamp:
        # Numpy's datetime format isn't compatible with TDMS,
        # so can't use data.tofile
        write_values(file, tdms_object.data)
    elif tdms_object.data_type == String:
        # Strings are variable size so need to be treated specially
        write_string_values(file, tdms_object.data)
    else:
        try:
            to_file(file, tdms_object.data)
        except (AttributeError):
            # Need to also handle lists of data,
            # to handle timestamp data for example.
            write_values(file, tdms_object.data)


def to_file(file, array):
    """Wrapper around ndarray.tofile to support any file-like object"""

    try:
        array.tofile(file)
    except (TypeError, IOError, UnsupportedOperation):
        file.write(array.tobytes())


def write_values(file, array):
    file.write(b''.join(_to_tdms_value(val).bytes for val in array))


def write_string_values(file, strings):
    try:
        encoded_strings = [s.encode("utf-8") for s in strings]
    except AttributeError:
        # Assume if we can't encode then we already have bytes
        encoded_strings = strings
    offset = 0
    for s in encoded_strings:
        offset += len(s)
        file.write(Uint32(offset).bytes)
    for s in encoded_strings:
        file.write(s)


def object_data_size(data_type, data_values):
    if data_type == String:
        # For string data, the total size is 8 bytes per string for the
        # offsets to the start of each string, plus the length of each string.
        try:
            encoded_strings = [s.encode("utf-8") for s in data_values]
        except AttributeError:
            encoded_strings = data_values
        return sum(4 + len(s) for s in encoded_strings)

    return data_type.size * len(data_values)


def _to_np_array(data):
    if isinstance(data, np.ndarray):
        return data

    dtype = _infer_dtype(data)
    return np.array(data, dtype=dtype)


def _infer_dtype(data):
    if data and isinstance(data[0], int):
        max_value = max(data)
        min_value = min(data)
        if max_value >= 2**63 and min_value >= 0:
            return np.dtype('uint64')
        elif max_value >= 2**32 or min_value < -1 * 2**31:
            return np.dtype('int64')
        elif max_value >= 2**31 and min_value >= 0:
            return np.dtype('uint32')
        elif max_value >= 2**16 or min_value < -1 * 2**15:
            return np.dtype('int32')
        elif max_value >= 2**15 and min_value >= 0:
            return np.dtype('uint16')
        elif max_value >= 2**8 or min_value < -1 * 2**7:
            return np.dtype('int16')
        elif max_value >= 2**7 and min_value >= 0:
            return np.dtype('uint8')
        else:
            return np.dtype('int8')
    return None



""" Lower level TDMS reader API that allows finer grained reading of data
"""

from collections import OrderedDict
import logging
import os
import numpy as np
import struct

from nptdms.common import toc_properties
from nptdms.utils import Timer
from nptdms.base_segment import RawChannelDataChunk
from nptdms.tdms_segment import TdmsSegment, SegmentIndexCache
from nptdms.log import log_manager


log = log_manager.get_logger(__name__)
_struct_unpack = struct.unpack


class TdmsReader(object):
    """ Reads metadata and data from a TDMS file.

    :ivar object_metadata: Dictionary of object path to ObjectMetadata
    """

    def __init__(self, tdms_file):
        """ Initialise a new TdmsReader

        :param tdms_file: Either the path to the tdms file to read
            as a string or pathlib.Path, or an already opened file.
        """
        self._file_path = None
        self._index_file_path = None
        self._file = None
        self._index_file = None

        self._segments = None
        self._prev_segment_objects = {}
        self.object_metadata = OrderedDict()
        self._segment_channel_offsets = {}
        self.tdms_version = None

        if hasattr(tdms_file, "read"):
            # Is a file
            tag = tdms_file.read(4)
            tdms_file.seek(0, os.SEEK_SET)

            if tag == b"TDSh":
                self._index_file = tdms_file
            elif tag == b"TDSm":
                self._file = tdms_file
            else:
                raise ValueError(
                    f"File should either start with 'b`TDSh`' or 'b`TDSm`', submitted starts with '{tag}'.")

        else:
            # Is path to a file
            source_path = str(tdms_file)
            if source_path.endswith(".tdms_index"):
                self._index_file_path = source_path
                self._index_file = open(self._index_file_path, "rb")

            else:
                self._file_path = source_path
                self._file = open(self._file_path, "rb")

                filepath = self._file_path + '_index'
                if os.path.isfile(filepath):
                    self._index_file_path = filepath
                    self._index_file = open(self._index_file_path, "rb")

    def close(self):
        if self._file is None and self._index_file is None:
            # Already closed
            return

        if self._file_path is not None:
            # File path was provided so we opened the file and should close it.
            self._file.close()

        if self._index_file_path is not None:
            # Index file path was provided so we opened the file and should close it.
            self._index_file.close()

        # Finally always remove reference to the files
        self._file = None
        self._index_file = None

    def read_metadata(self, require_segment_indexes=False):
        """ Read all metadata and structure information from a TdmsFile

        :param require_segment_indexes: Whether to create segment object indexes to allow lookup of objects by path.
        """
        if self._index_file is not None:  # generally try to read metadata from index file because it is faster
            file = self._index_file
            reading_index_file = True
        elif self._file is not None:  # fallback if only a data file is supplied
            file = self._file
            reading_index_file = False
        else:
            raise ValueError("Neither tdms_index file nor tdms file is available.")

        self._segments = []
        segment_position = 0
        try:
            with Timer(log, "Read metadata"):
                # Read metadata first to work out how much space we need
                previous_segment = None
                index_cache = SegmentIndexCache() if require_segment_indexes else None
                while True:
                    start_position = file.tell()
                    try:
                        segment, properties = self._read_segment_metadata(
                            file, segment_position, index_cache, previous_segment, reading_index_file)
                    except EOFError:
                        # We've finished reading the file
                        break

                    self._update_object_metadata(segment)
                    self._update_object_properties(properties)
                    self._segments.append(segment)
                    previous_segment = segment

                    segment_position = segment.next_segment_pos
                    if reading_index_file:
                        file.seek(start_position + segment.data_position - segment.position, os.SEEK_SET)
                    else:
                        file.seek(segment.next_segment_pos, os.SEEK_SET)
        finally:
            if reading_index_file and self._index_file_path is not None:
                file.close()

    def read_raw_data(self):
        """ Read raw data from all segments, chunk by chunk

        :returns: A generator that yields RawDataChunk objects
        """
        self._ensure_open()
        if self._segments is None:
            raise RuntimeError(
                "Cannot read data unless metadata has first been read")
        for segment in self._segments:
            self._verify_segment_start(segment)
            for chunk in segment.read_raw_data(self._file):
                yield chunk

    def read_raw_data_for_channel(self, channel_path, offset=0, length=None):
        """ Read raw data for a single channel, chunk by chunk

        :param channel_path: The path of the channel object to read data for
        :param offset: Initial position to read data from.
        :param length: Number of values to attempt to read.
            If None, then all values starting from the offset will be read.
            Fewer values will be returned if attempting to read beyond the end of the available data.
        :returns: A generator that yields RawChannelDataChunk objects
        """
        self._ensure_open()
        if self._segments is None:
            raise RuntimeError("Cannot read data unless metadata has first been read")

        try:
            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]
        except KeyError:
            with Timer(log, "Build data index for channel"):
                self._build_index(channel_path)
            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]

        object_metadata = self.object_metadata[channel_path]
        max_length_from_offset = object_metadata.num_values - offset
        if length is None:
            length = max_length_from_offset
        else:
            # Make sure we're not trying to read more data than is actually available
            length = min(length, max_length_from_offset)
        end_index = offset + length

        # Binary search to find first and last segments to read
        start_segment = first_segment + np.searchsorted(segment_offsets, offset, side='right')
        end_segment = first_segment + np.searchsorted(segment_offsets, end_index, side='left')

        segment_index = start_segment
        values_read = 0
        for segment in self._segments[start_segment:end_segment + 1]:
            self._verify_segment_start(segment)
            # By default, read all chunks in a segment
            chunk_offset = 0
            num_chunks = segment.num_chunks
            segment_obj = segment.get_segment_object(channel_path)
            chunk_size = 0 if (segment_obj is None or not segment_obj.has_data) else segment_obj.number_values
            if chunk_size == 0:
                continue
            segment_start_index = (
                0 if segment_index == first_segment else segment_offsets[segment_index - first_segment - 1])
            remaining_values_to_skip = 0

            # For the first and last segments, we may not need to read all chunks,
            # and may need to trim some data from the beginning or end of the chunk.
            if segment_index == start_segment:
                num_values_to_skip = offset - segment_start_index
                chunk_offset = num_values_to_skip // chunk_size
                remaining_values_to_skip = num_values_to_skip % chunk_size
                num_chunks -= chunk_offset
            if segment_index == end_segment:
                # Note: segment_index may be both start and end
                segment_end_index = segment_offsets[segment_index - first_segment]
                num_values_to_trim = segment_end_index - end_index

                # Account for segments where the final chunk is truncated
                final_chunk_size = (segment_end_index - segment_start_index) % chunk_size
                final_chunk_size = chunk_size if final_chunk_size == 0 else final_chunk_size
                if num_values_to_trim >= final_chunk_size:
                    num_chunks -= 1
                    num_values_to_trim -= final_chunk_size

                num_chunks -= num_values_to_trim // chunk_size

            for i, chunk in enumerate(
                    segment.read_raw_data_for_channel(self._file, channel_path, chunk_offset, num_chunks)):
                skip = remaining_values_to_skip if i == 0 else 0
                values_read += len(chunk) - skip
                trim = 0 if values_read < length else values_read - length
                yield _trim_channel_chunk(chunk, skip, trim)

            segment_index += 1

    def read_channel_chunk_for_index(self, channel_path, index):
        """ Read the chunk containing the given index

        :returns: Tuple of raw channel data chunk and the integer offset to the beginning of the chunk
        :rtype: (RawChannelDataChunk, int)
        """
        self._ensure_open()
        if self._segments is None:
            raise RuntimeError("Cannot read data unless metadata has first been read")

        try:
            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]
        except KeyError:
            with Timer(log, "Build data index for channel"):
                self._build_index(channel_path)
            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]

        # Binary search to find the segment to read
        segment_index = first_segment + np.searchsorted(segment_offsets, index, side='right')
        segment = self._segments[segment_index]
        segment_obj = segment.get_segment_object(channel_path)
        chunk_size = 0 if segment_obj is None else segment_obj.number_values
        segment_start_index = (
            0 if segment_index == first_segment else segment_offsets[segment_index - first_segment - 1])

        index_in_segment = index - segment_start_index
        chunk_index = index_in_segment // chunk_size

        self._verify_segment_start(segment)
        chunk_data = next(segment.read_raw_data_for_channel(self._file, channel_path, chunk_index, 1))
        chunk_offset = segment_start_index + chunk_index * chunk_size
        return chunk_data, chunk_offset

    def is_index_file_only(self):
        """ Convenience function to access if the supplied file is an index file and no data file is available

        :rtype: bool
        """
        return self._file is None and self._index_file is not None

    def _read_segment_metadata(
            self, file, segment_position, index_cache, previous_segment, is_index_file):
        (position, toc_mask, data_position, next_segment_pos, segment_incomplete) = self._read_lead_in(
            file, segment_position, is_index_file)

        segment = TdmsSegment(
            position, toc_mask, next_segment_pos, data_position)

        properties = segment.read_segment_objects(
            file, self._prev_segment_objects, index_cache, previous_segment, segment_incomplete)
        return segment, properties

    def _read_lead_in(self, file, segment_position, is_index_file=False):
        lead_in_bytes = file.read(28)
        if len(lead_in_bytes) < 28:
            raise EOFError

        expected_tag = b'TDSh' if is_index_file else b'TDSm'
        tag = lead_in_bytes[:4]
        if tag != expected_tag:
            raise ValueError(
                "Segment does not start with %r, but with %r" % (expected_tag, tag))

        # Next four bytes are table of contents mask
        toc_mask = _struct_unpack('<l', lead_in_bytes[4:8])[0]

        if log.isEnabledFor(logging.DEBUG):
            log.debug("Reading segment at %d", segment_position)
            for prop_name, prop_mask in toc_properties.items():
                prop_is_set = (toc_mask & prop_mask) != 0
                log.debug("Property %s is %s", prop_name, prop_is_set)

        endianness = '>' if (toc_mask & toc_properties['kTocBigEndian']) else '<'

        # Next four bytes are version number, then 8 bytes each for the offset values
        (version, next_segment_offset, raw_data_offset) = _struct_unpack(endianness + 'lQQ', lead_in_bytes[8:28])

        if self.tdms_version is None:
            if version not in (4712, 4713):
                log.warning("Unrecognised version number: %d" % version)
            self.tdms_version = version
        elif self.tdms_version != version:
            log.warning("Segment version mismatch, %d != %d" % (version, self.tdms_version))

        # Calculate data and next segment position
        lead_size = 7 * 4
        data_position = segment_position + lead_size + raw_data_offset
        segment_incomplete = next_segment_offset == 0xFFFFFFFFFFFFFFFF
        if segment_incomplete:
            # Segment size is unknown. This can happen if LabVIEW crashes.
            next_segment_pos = self._get_data_file_size()
            if next_segment_pos < data_position:
                # Metadata wasn't completely written and don't have any data in this segment,
                # don't try to read any metadata
                log.warning("Last segment metadata is incomplete")
                raise EOFError
            # Try to read until the end of the file if we have complete metadata
            log.warning(
                "Last segment of file has unknown size, "
                "will attempt to read to the end of the file")
        else:
            log.debug("Next segment offset = %d, raw data offset = %d, data size = %d b",
                      next_segment_offset, raw_data_offset, next_segment_offset - raw_data_offset)
            next_segment_pos = (
                    segment_position + next_segment_offset + lead_size)

        return segment_position, toc_mask, data_position, next_segment_pos, segment_incomplete

    def _verify_segment_start(self, segment):
        """ When reading data for a segment, check for the TDSm tag at the start of the segment in an attempt
            to detect any mismatch between tdms and tdms_index files.
        """
        position = segment.position
        self._file.seek(segment.position)
        expected_tag = b'TDSm'
        tag = self._file.read(4)
        if tag != expected_tag:
            raise ValueError(
                "Attempted to read data segment at position {0} but did not find segment start header. ".format(
                    position) +
                "Check that the tdms_index file matches the tdms data file.")

    def _get_data_file_size(self):
        current_pos = self._file.tell()
        self._file.seek(0, os.SEEK_END)
        end_pos = self._file.tell()
        self._file.seek(current_pos, os.SEEK_SET)
        return end_pos

    def _update_object_metadata(self, segment):
        """ Update object metadata using the metadata read from a single segment
        """
        for segment_object in segment.ordered_objects:
            path = segment_object.path
            self._prev_segment_objects[path] = segment_object

            object_metadata = self._get_or_create_object(path)
            object_metadata.num_values += _number_of_segment_values(segment_object, segment)
            _update_object_data_type(path, object_metadata, segment_object)
            if segment_object.scaler_data_types is not None:
                _update_object_scaler_data_types(path, object_metadata, segment_object)

    def _update_object_properties(self, segment_object_properties):
        """ Update object properties using any properties in a segment
        """
        if segment_object_properties is not None:
            for path, properties in segment_object_properties.items():
                object_metadata = self._get_or_create_object(path)
                for prop, val in properties:
                    object_metadata.properties[prop] = val

    def _get_or_create_object(self, path):
        """ Get existing object metadata or create metadata for a new object
        """
        try:
            return self.object_metadata[path]
        except KeyError:
            obj = ObjectMetadata()
            self.object_metadata[path] = obj
            return obj

    def _build_index(self, channel_path):
        """ Builds an index into the segment data for faster lookup of values

            _segment_channel_offsets provides data offset at the end of each segment per channel
        """
        num_segments = len(self._segments)

        # Get number of values for this channel in each segment
        segment_num_values = np.zeros(num_segments, dtype=np.int64)
        first_segment = -1
        last_segment = -1

        for i, segment in enumerate(self._segments):
            obj_index = segment.object_index.get(channel_path)
            if obj_index is not None:
                segment_obj = segment.ordered_objects[obj_index]
                num_values = _number_of_segment_values(segment_obj, segment)
                if num_values > 0:
                    segment_num_values[i] = num_values
                    last_segment = i
                    if first_segment == -1:
                        first_segment = i

        # Now use the cumulative sum to get the total channel value count
        # at the end of each segment.
        if first_segment == -1:
            first_segment = num_segments
            last_segment = num_segments
        channel_offsets = np.cumsum(segment_num_values[first_segment:last_segment + 1])

        # It's likely that many channels will have the same shaped data,
        # so de-duplicate these arrays to reduce memory usage.
        existing_arrays = (xs for (_, xs) in self._segment_channel_offsets.values())
        channel_offsets = _deduplicate_array(channel_offsets, existing_arrays)
        self._segment_channel_offsets[channel_path] = (first_segment, channel_offsets)

    def _ensure_open(self):
        if self._file is None and self._index_file is None:
            raise RuntimeError(
                "Cannot read data after the underlying TDMS reader is closed")


def _number_of_segment_values(segment_object, segment):
    """ Compute the number of values an object has in a segment
    """
    if not segment_object.has_data:
        return 0
    if segment.final_chunk_lengths_override is None:
        return segment_object.number_values * segment.num_chunks
    else:
        return (segment_object.number_values * (segment.num_chunks - 1) +
                segment.final_chunk_lengths_override.get(segment_object.path, 0))


def _update_object_data_type(path, obj, segment_object):
    """ Update the data type for an object using its segment metadata
    """
    if obj.data_type is not None and obj.data_type != segment_object.data_type:
        raise ValueError(
            "Segment data doesn't have the same type as previous "
            "segments for objects %s. Expected type %s but got %s" %
            (path, obj.data_type, segment_object.data_type))
    obj.data_type = segment_object.data_type


def _update_object_scaler_data_types(path, obj, segment_object):
    """ Update the DAQmx scaler data types for an object using its segment metadata
    """
    if obj.scaler_data_types is not None and obj.scaler_data_types != segment_object.scaler_data_types:
        raise ValueError(
            "Segment data doesn't have the same scaler data types as previous "
            "segments for objects %s. Expected types %s but got %s" %
            (path, obj.scaler_data_types, segment_object.scaler_data_types))
    obj.scaler_data_types = segment_object.scaler_data_types


class ObjectMetadata(object):
    """ Stores information about an object in a TDMS file
    """
    def __init__(self):
        self.properties = OrderedDict()
        self.data_type = None
        self.scaler_data_types = None
        self.num_values = 0


def _trim_channel_chunk(chunk, skip=0, trim=0):
    if skip == 0 and trim == 0:
        return chunk
    data = None
    scaler_data = None
    if chunk.data is not None:
        data = chunk.data[skip:len(chunk.data) - trim]
    if chunk.scaler_data is not None:
        scaler_data = {
            scale_id: d[skip:len(d) - trim]
            for (scale_id, d) in chunk.scaler_data.items()}
    return RawChannelDataChunk(data, scaler_data)


def _deduplicate_array(xs, candidates):
    """ Reduce memory usage by replacing an array with a reference to an existing array if equal
    """
    for candidate in candidates:
        if _array_equal(xs, candidate):
            return candidate
    return xs


def _array_equal(a, b, chunk_size=100):
    """ Compare two arrays for equality
    """
    # Numpy array_equal compares all elements rather than comparing one at a time and short-circuiting when it
    # finds a difference. Break up the comparison into chunks to make this faster. Adapted from:
    # https://stackoverflow.com/questions/26260848/numpy-fast-check-for-complete-array-equality-like-matlabs-isequal
    if len(a) != len(b):
        return False

    num_chunks = (len(a) + chunk_size - 1) // chunk_size
    for i in range(num_chunks):
        offset = i * chunk_size
        if not (a[offset:offset+chunk_size] == b[offset:offset+chunk_size]).all():
            return False
    return True



from __future__ import print_function

from argparse import ArgumentParser
import logging

from nptdms import TdmsFile
from nptdms.log import log_manager


def main():
    parser = ArgumentParser(
        description="List the contents of a LabView TDMS file.")
    parser.add_argument(
        '-p', '--properties', action="store_true",
        help="Include channel properties.")
    parser.add_argument(
        '-d', '--debug', action="store_true",
        help="Print debugging information to stderr.")
    parser.add_argument(
        'tdms_file',
        help="TDMS file to read.")
    args = parser.parse_args()

    if args.debug:
        log_manager.set_level(logging.DEBUG)

    tdmsinfo(args.tdms_file, args.properties)


def tdmsinfo(file, show_properties=False):
    tdms_file = TdmsFile.read_metadata(file)

    level = 0
    display('/', level)
    if show_properties:
        display_properties(tdms_file, level + 1)
    for group in tdms_file.groups():
        level = 1
        display("%s" % group.path, level)
        if show_properties:
            display_properties(group, level + 1)
        for channel in group.channels():
            level = 2
            display("%s" % channel.path, level)
            if show_properties:
                level = 3
                if channel.data_type is not None:
                    display("data type: %s" % channel.data_type.__name__, level)
                display("length: %d" % len(channel), level)
                display_properties(channel, level)


def display_properties(tdms_object, level):
    if tdms_object.properties:
        display("properties:", level)
        for prop, val in tdms_object.properties.items():
            display("%s: %s" % (prop, val), level + 1)


def display(s, level):
    print("%s%s" % (" " * 2 * level, s))
